#import "PendoFlutterPlugin.h"
#import <Pendo/Pendo.h>

typedef NS_ENUM(NSInteger, METHOD_CALLS_OPCODES) {
    UNAVAILABLE              = -1,
    SET_VISITOR_DATA         = 1,
    SET_ACCOUNT_DATA         = 2,
    TRACK                    = 3,
    GET_VISITOR_ID           = 4,
    GET_ACCOUNT_ID           = 5,
    GET_DEVICE_ID            = 6,
    END_SESSION              = 7,
    PAUSE_GUIDES             = 8,
    RESUME_GUIDES            = 9,
    DISMISS_VISIBLE_GUIDES   = 10,
    SET_DEBUG_MODE           = 11,
    INIT_SDK_SETUP           = 12,
    START_SESSION            = 13,
    SCREEN_CHANGED           = 14
};

static NSString *const kPNDPendoChannelName = @"pendo_sdk_channel_name";
static NSString *const kPNDPendoPluginName = @"pendo_sdk";
static NSString *const kIntegrationType = @"IntegrationType";
static NSString *const kPendoOptions = @"PendoOptions";
static NSString *const kApplicationKey = @"ApplicationKey";
static NSString *const kPlatformVersion = @"PlatformVersion";
static NSString *const kVisitorID = @"VisitorID";
static NSString *const kAccountID = @"AccountID";
static NSString *const kVisitorData = @"VisitorData";
static NSString *const kAccountData = @"AccountData";
static NSString *const kIsDebugEnabled = @"IsDebugEnabled";
static NSString *const kDismissGuides = @"DismissGuides";
static NSString *const kTrackEventName = @"EventName";
static NSString *const kTrackEventProperties = @"EventProperties";
static NSString *const kTrack = @"Track";
static NSString *const kRetroactiveScreenId = @"retroactiveScreenId";
static NSString *const kClickableElements = @"clickableElements";
static NSString *const kEnvironmentName = @"environmentName";
static NSString *const kConfigs = @"configs";
static NSString *const kInfo = @"info";

@implementation PendoFlutterPlugin
+ (void)registerWithRegistrar:(NSObject<FlutterPluginRegistrar>*)registrar {
    FlutterMethodChannel* channel = [FlutterMethodChannel methodChannelWithName:kPNDPendoChannelName binaryMessenger:[registrar messenger]];
    PendoFlutterPlugin* instance = [[PendoFlutterPlugin alloc] init];
    [registrar addMethodCallDelegate:instance channel:channel];
}

+ (void)registerWithRegistry:(NSObject<FlutterPluginRegistry>*)registry {
    NSObject<FlutterPluginRegistrar> *registrar = [registry registrarForPlugin:kPNDPendoPluginName];
    [self registerWithRegistrar:registrar];
}

- (void)handleMethodCall:(FlutterMethodCall *)call result:(FlutterResult)result {
    NSLog(@"PendoFlutterPlugin method name: %@", call.method);
    NSUInteger methodOpcode = [self methodOpcode:call.method];
    id arguments = call.arguments;

    switch (methodOpcode) {
        case SET_VISITOR_DATA: {
            [self setVisitorData:arguments result:result];
            break;
        } case TRACK: {
            [self track:arguments result:result];
            break;
        } case GET_VISITOR_ID: {
            [self getVisitorId:arguments result:result];
            break;
        } case GET_ACCOUNT_ID: {
            [self getAccountId:arguments result:result];
            break;
        } case GET_DEVICE_ID: {
            [self getDeviceId:arguments result:result];
            break;
        } case SET_ACCOUNT_DATA: {
            [self setAccountData:arguments result:result];
            break;
        } case END_SESSION: {
            [self endSession:arguments result:result];
            break;
        } case PAUSE_GUIDES: {
            [self pauseGuides:arguments result:result];
            break;
        } case RESUME_GUIDES: {
            [self resumeGuides:arguments result:result];
            break;
        } case DISMISS_VISIBLE_GUIDES: {
            [self dismissVisibleGuides:arguments result:result];
            break;
        } case SET_DEBUG_MODE: {
            [self setDebugMode:arguments result:result];
            break;
        } case INIT_SDK_SETUP: {
            [self setup:arguments result:result];
            break;
        } case START_SESSION: {
            [self startSession:arguments result:result];
            break;
        } case SCREEN_CHANGED: {
            [self screenChanged:arguments result:result];
            break;
        } default: {
            [self methodNotImplemented:arguments result:result];
            break;
        }
    }
}

#pragma mark - SDK Wrappers
- (void)setup:(id)args result:(FlutterResult)result {
    if (nil != args && [args isKindOfClass:[NSDictionary class]]) {
        NSDictionary *arguments = (NSDictionary *)args;
        NSString *appKey = arguments[kApplicationKey];
        PendoOptions *options = [self optionsFromArgs:arguments];
        [self setup:appKey withOptions:options];
        result(nil);
    } else {
        FlutterError *error = [FlutterError errorWithCode:@(INIT_SDK_SETUP).stringValue message:@"setup Call Failed" details:nil];
        result(error);
    }
}

- (void)startSession:(id)args result:(FlutterResult)result {
    if (nil != args && [args isKindOfClass:[NSDictionary class]]) {
        NSDictionary *arguments = args;
        NSString *visitorId = arguments[kVisitorID];
        if ([visitorId isKindOfClass:[NSNull class]]) {
            visitorId = nil;
        }
        NSString *accountId = arguments[kAccountID];
        if ([accountId isKindOfClass:[NSNull class]]) {
            accountId = nil;
        }
        NSDictionary *visitorData = arguments[kVisitorData];
        if ([visitorData isKindOfClass:[NSNull class]]) {
            visitorData = nil;
        }
        NSDictionary *accountData = arguments[kAccountData];
        if ([accountData isKindOfClass:[NSNull class]]) {
            accountData = nil;
        }
        [self startSession:visitorId accountId:accountId visitorData:visitorData accountData:accountData];
        result(nil);
    } else {
        FlutterError *error = [FlutterError errorWithCode:@(START_SESSION).stringValue message:@"startSession Call Failed" details:nil];
        result(error);
    }
}

- (void)setVisitorData:(id)args result:(FlutterResult)result {
    if (nil != args && [args isKindOfClass:[NSDictionary class]]) {
        NSDictionary *arguments = (NSDictionary *)args;
        NSDictionary *visitorData = arguments[kVisitorData];
        [self setVisitorData:visitorData];
        result(nil);
    } else  {
        FlutterError *error = [FlutterError errorWithCode:@(SET_VISITOR_DATA).stringValue message:@"SetVisitorData Call Failed" details:nil];
        result(error);
    }
}

- (void)setAccountData:(id)args result:(FlutterResult)result {
    if (nil != args && [args isKindOfClass:[NSDictionary class]]) {
        NSDictionary *arguments = (NSDictionary *)args;
        NSDictionary *accountData = arguments[kAccountData];
        [self setAccountData:accountData];
        result(nil);
    } else {
        FlutterError *error = [FlutterError errorWithCode:@(SET_ACCOUNT_DATA).stringValue message:@"setAccountData Call Failed" details:nil];
        result(error);
    }
}

- (void)track:(id)args result:(FlutterResult)result {
    if (nil != args && [args isKindOfClass:[NSDictionary class]]) {
        NSDictionary *arguments = (NSDictionary *)args;
        NSString *eventName = arguments[kTrackEventName];
        NSDictionary *properties = arguments[kTrackEventProperties];
        [self track:eventName properties:properties];
        result(nil);
    } else {
        FlutterError *error = [FlutterError errorWithCode:@(TRACK).stringValue message:@"Track Call Failed" details:nil];
        result(error);
    }
}

- (void)getVisitorId:(id)args result:(FlutterResult)result {
    result([self visitorId]);
}

- (void)getAccountId:(id)args result:(FlutterResult)result {
    result([self accountId]);
}

- (void)getDeviceId:(id)args result:(FlutterResult)result {
    result([self deviceId]);
}

- (void)endSession:(id)args result:(FlutterResult)result {
    [self endSession];
    result(nil);
}

- (void)pauseGuides:(id)args result:(FlutterResult)result {
    if (nil != args && [args isKindOfClass:[NSDictionary class]]) {
        NSDictionary *arguments = (NSDictionary *)args;
        BOOL shouldDismissGuides = [arguments[kDismissGuides] boolValue];
        [self pauseGuides:shouldDismissGuides];
        result(nil);
    } else {
        FlutterError *error = [FlutterError errorWithCode:@(PAUSE_GUIDES).stringValue message:@"PauseGuides Call Failed" details:nil];
        result(error);
    }
}

- (void)resumeGuides:(id)args result:(FlutterResult)result {
    [self resumeGuides];
    result(nil);
}

- (void)dismissVisibleGuides:(id)args result:(FlutterResult)result {
    [self dismissVisibleGuides];
    result(nil);
}

- (void)setDebugMode:(id)args result:(FlutterResult)result {
    if (nil != args && [args isKindOfClass:[NSDictionary class]]) {
        NSDictionary *arguments = (NSDictionary *)args;
        BOOL isDebugEnabled = [arguments[kIsDebugEnabled] boolValue];
        [self setDebugMode:isDebugEnabled];
        result(nil);
    } else {
        FlutterError *error = [FlutterError errorWithCode:@(SET_DEBUG_MODE).stringValue message:@"SetDebugMode Call Failed" details:nil];
        result(error);
    }
}

- (void)methodNotImplemented:(id)args result:(FlutterResult) result {
    result(FlutterMethodNotImplemented);
}

#pragma mark - SDK methods
/**
 * Must be called <b>after</b> the SDK was initialized.
 *
 * @brief start a session with a new visitor.
 * In case a session was already started before, end the previous one and start a new one.
 *
 * @param visitorId The visitor's ID.
 * @param accountId The account's ID.
 * @param visitorData The visitor's data.
 * @param accountData The account's data.
 */
- (void)startSession:(nullable NSString *)visitorId
        accountId:(nullable NSString *)accountId
        visitorData:(nullable NSDictionary *)visitorData
        accountData:(nullable NSDictionary *)accountData {
            [[PendoManager sharedManager] startSession:visitorId accountId:accountId visitorData:[self validateDictionaryParam:visitorData] accountData:[self validateDictionaryParam:accountData]];
}

/**
 * Set a visitor data value.
 * This data is used by Pendo Mobile for creating audiences or reporting analytics.
 * For instance you might want to provide data on the visitor's age or if the visitor is logged into a service.
 *
 *  @param visitorData additional visitor data.
 */
- (void)setVisitorData:(nonnull NSDictionary *)visitorData {
    [[PendoManager sharedManager] setVisitorData:[self validateDictionaryParam:visitorData]];
}

/**
 * Set account data value for a given data name.
 * This data is used by Pendo Mobile for creating audiences or reporting analytics.
 * For instance you might want to provide data on the account's subscription or if the account is active or not.
 *
 * @param accountData the account data.
 */
- (void)setAccountData:(nonnull NSDictionary *)accountData {
    [[PendoManager sharedManager] setAccountData:[self validateDictionaryParam:accountData]];
}

/**
 * When your application needs to send additional events about actions your users perform.
 *
 * @param event The event name describing the user’s action.
 * @param properties dictionary of event properties (optional).
 * @brief Queue a track eventsfor transmission, optionally including properties as the payload of the event.
 */
- (void)track:(nonnull NSString *)event properties:(nullable NSDictionary *)properties {
    [[PendoManager sharedManager] track:event properties:[self validateDictionaryParam:properties]];
}

/**
 *  Call this method on the sharedManger with your application key.
 *  Setup the Pendo SDK. Call startSession when you would like to start the session (anonymous or identified)
 *
 *  @param appKey The app key for your account
 */
- (void)setup:(NSString *_Nonnull)appKey {
    [[PendoManager sharedManager] setup:appKey];
}

/**
 *  Call this method on the sharedManger with your application key.
 *  Setup the Pendo SDK. Call startSession when you would like to start the session (anonymous or identified)
 *
 *  @param appKey The app key for your account
 *  @param options additional options for internal use only, default nil
 */
- (void)setup:(NSString *)appKey withOptions:(PendoOptions *_Nullable)options {
    [[PendoManager sharedManager] setup:appKey withOptions:options];
}

/**
 *  Provide a visitor id to the Pendo Mobile SDK.
 */
- (NSString *)visitorId {
    return [[PendoManager sharedManager] visitorId];
}

/**
 *  Provide a account id to the Pendo Mobile SDK.
 */
- (NSString *)accountId {
    return [[PendoManager sharedManager] accountId];
}

/**
 * Get Pendo SDK unique device id. Used for anonymous visitor id.
 * This id is unique per application.
 */
- (NSString *)deviceId {
    return [[PendoManager sharedManager] getDeviceId];
}

/**
 *  Call this method on the sharedManger to end current session and stop gathering analytics.
 *  @warning Must be called <b>after</b> the SDK was initialized.
 */
- (void)endSession {
    [[PendoManager sharedManager] endSession];
}

/**
 * Call in order to stop showing guides.
 * @param dismissGuides indicates if to dismiss on going guides from screen or not
 */
- (void)pauseGuides:(BOOL)dismissGuides {
    [[PendoManager sharedManager] pauseGuides:dismissGuides];
}

/**
 * Call in order to resume showing guides.
 */
- (void)resumeGuides {
    [[PendoManager sharedManager] resumeGuides];
}

/**
 * Dismiss any visible guides
 */
- (void)dismissVisibleGuides {
    [[PendoManager sharedManager] dismissVisibleGuides];
}

/**
 * Set debug mode to true to display log messages
 */
- (void)setDebugMode:(BOOL)isDebugEnabled {
    [[PendoManager sharedManager] setDebugMode:isDebugEnabled];
}

- (void)screenChanged:(id)args result:(FlutterResult)result {
    if (nil != args && [args isKindOfClass:[NSDictionary class]]) {
        NSDictionary *arguments = (NSDictionary *)args;
        NSString *screenName = arguments[kRetroactiveScreenId];
        NSDictionary *info = arguments[kInfo];
        NSLog(@"screen name: %@ info: %@", screenName, info);
        [PendoManagerFlutter screenChanged:screenName info:info];
        result(nil);
    } else {
        FlutterError *error = [FlutterError errorWithCode:@(SCREEN_CHANGED).stringValue message:@"screenChanged Call Failed" details:nil];
        result(error);
    }
}

- (NSInteger)methodOpcode:(NSString *)methodName {
    if (methodName.length == 0) {
        return UNAVAILABLE;
    }
    NSDictionary <NSString*, NSNumber *> *methodNamesOpcode = [self methodNamesOpcodes];
    return methodNamesOpcode[methodName].intValue ?: UNAVAILABLE;
}

- (NSDictionary <NSString*, NSNumber *>*)methodNamesOpcodes {
    static NSDictionary *methodNames = nil;
    static dispatch_once_t once;
    dispatch_once(&once,^{
        methodNames =  @{
            @"setVisitorData":         @(SET_VISITOR_DATA),
            @"setAccountData":         @(SET_ACCOUNT_DATA),
            @"track":                  @(TRACK),
            @"getVisitorId":           @(GET_VISITOR_ID),
            @"getAccountId":           @(GET_ACCOUNT_ID),
            @"getDeviceId":            @(GET_DEVICE_ID),
            @"endSession":             @(END_SESSION),
            @"pauseGuides":            @(PAUSE_GUIDES),
            @"resumeGuides":           @(RESUME_GUIDES),
            @"dismissVisibleGuides":   @(DISMISS_VISIBLE_GUIDES),
            @"setDebugMode":           @(SET_DEBUG_MODE),
            @"setup":                  @(INIT_SDK_SETUP),
            @"startSession":           @(START_SESSION),
            @"screenChanged":          @(SCREEN_CHANGED)
        };
    });

    return methodNames;
}

#pragma mark - Private
- (NSDictionary *_Nullable)validateDictionaryParam:(NSDictionary *_Nullable)input {
    if (input == nil || [input isKindOfClass:[NSNull class]]) {
        return nil;
    }

    NSMutableDictionary *result = [NSMutableDictionary dictionary];

    for (id key in input.allKeys) {
        if (![input[key] isKindOfClass:[NSNull class]]) {
            [result setObject:input[key] forKey:key];
        } else {
            NSLog(@"removing value for key %@ since its NSNull", key);
        }
    }

    return [result copy];
}

- (PendoOptions *)optionsFromArgs:(NSDictionary *)arguments {
    NSDictionary *sdkArgs = arguments[kPendoOptions];
    NSString *platfromVersion = arguments[kPlatformVersion];
    PendoOptions *options = [[PendoOptions alloc] init];
    NSString *integrationType = kTrack;
    
    if (![sdkArgs isKindOfClass:[NSNull class]]) {
        NSString *environmentName = sdkArgs[kEnvironmentName];
        NSDictionary *configs = sdkArgs[kConfigs];
        options.environmentName = environmentName;
        options.configs = [configs mutableCopy];
    
        if (sdkArgs[kIntegrationType] != nil) {
            integrationType = sdkArgs[kIntegrationType];
        }
    }
    
    [options setFlutterIntegrationType:integrationType];
    [options setIsFlutter:YES];
    [options setPlatformVersion:platfromVersion];
    
    return options;
}

@end
