import '../logger.dart';

extension MapValidator on Map {

  Map validateData() {
    Map result = {};
    final keys = this.keys;
    Object? newValue;

    for (final key in keys) {
      newValue = _validateObject(this[key]); 

      if (newValue != null) {
        result[key] = newValue;
      }
    }

    return result;
  }

  Object? _validateObject(Object? input) {
    if (input == null) {
      return null;
    }
    
    Object? result = input;

    if (input is List) {
      result = _validateList(input);
    } else if (input is Set) {
      result = _validateSet(input);
    } else if (input is Map) {
      Logger.w('Map is not a supported as nested data type');
      result = null;
    } else if (input is DateTime) {
      result = input.toIso8601String();
    } else if (input is int) {
      // supported, return
    } else if (input is double) {
      // supported, return
    } else if (input is String) {
      // supported, return
    } else if (input is bool) {
      // supported, return
    } else {
      Logger.w('Custom objects are not a supported');
      result = null;
    } 

    return result;
  }

  List? _validateList(List list) {
    List result = [];

    for (final obj in list) {
      Object? checkedObj = _validateObject(obj);

      if (checkedObj != null) {
        result.add(checkedObj);
      } else {
        Logger.w('could not add object ' + obj.toString());
      }
    }

    return result;
  }

  Set? _validateSet(Set set) {
    Set result = {};

    for (final obj in set) {
      Object? checkedObj = _validateObject(obj);

      if (checkedObj != null) {
        result.add(checkedObj);
      } else {
        Logger.w('could not add object ' + obj.toString());
      }
    }

    return result;
  }
}
