/*
 *  * Copyright © 2019 Pendo.io. All rights reserved.
 */

// TODO: enhance Logger
enum _Level {
  trace,
  debug,
  info,
  warning,
  error,
  critical,
  fatal
}

final Map<_Level, String> _defaultLevelEmojis = {
  _Level.trace: '',
  _Level.debug: '✅',
  _Level.info: 'ℹ️',
  _Level.warning: '⚠️',
  _Level.critical: '❗️',
  _Level.error: '⛔',
  _Level.fatal: '💀',
};

class Logger {
  /// sets the debug level print. use [_Level.trace] to print all messages/release.
  static _Level _debugLevel = _Level.warning;
  
  static debugEnabled() {
    _debugLevel = _Level.trace;
  }

  /// API
  static void t(String msg) {
    _log(msg, _Level.trace);
  }

  static void d(String msg) {
    _log(msg, _Level.debug);
  }

  static void i(String msg) {
    _log(msg, _Level.info);
  }

  static void w(String msg) {
    _log(msg, _Level.warning);
  }

  static void c(String msg) {
    _log(msg, _Level.critical);
  }

  static void e(String msg) {
    _log(msg, _Level.error);
  }

  static void f(String msg) {
    _log(msg, _Level.fatal);
  }

  /// Private
  static _log(String msg, _Level lvl) {
    if (lvl.index >= _debugLevel.index) {
      String emoji = _defaultLevelEmojis[lvl] ?? '';
      String output = '[Pendo] ' + '[' + _getTime() + '] ' + emoji + ' '+ msg;
      print(output);
    }
  }

  static String _getTime({DateTime? time}) {
    if (time == null) {
      time = DateTime.now();
    }

    String threeDigits(int n) {
      if (n >= 100) return '$n';
      if (n >= 10) return '0$n';
      return '00$n';
    }

    String twoDigits(int n) {
      if (n >= 10) return '$n';
      return '0$n';
    }

    var now = time;
    var h = twoDigits(now.hour);
    var min = twoDigits(now.minute);
    var sec = twoDigits(now.second);
    var ms = threeDigits(now.millisecond);
    return '$h:$min:$sec.$ms';
  }
}